@@ -1,14 +1,16 @@
-# Copyright (C) 2016-present the ayncpg authors and contributors
+# Copyright (C) 2016-present the asyncpg authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of asyncpg and is released under
 # the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0
 
 
-from .connection import connect  # NOQA
+from .connection import connect, Connection  # NOQA
 from .exceptions import *  # NOQA
 from .pool import create_pool  # NOQA
+from .protocol import Record  # NOQA
 from .types import *  # NOQA
 
 
-__all__ = ('connect', 'create_pool') + exceptions.__all__  # NOQA
+__all__ = ('connect', 'create_pool', 'Record', 'Connection') + \
+          exceptions.__all__  # NOQA
