@@ -1,4 +1,4 @@
-# Copyright (C) 2016-present the ayncpg authors and contributors
+# Copyright (C) 2016-present the asyncpg authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of asyncpg and is released under
@@ -15,6 +15,7 @@ import re
 import shutil
 import socket
 import subprocess
+import sys
 import tempfile
 import textwrap
 import time
@@ -178,6 +179,15 @@ class Cluster:
         if sockdir is None:
             sockdir = '/tmp'
 
+        ssl_key = server_settings.get('ssl_key_file')
+        if ssl_key:
+            # Make sure server certificate key file has correct permissions.
+            keyfile = os.path.join(self._data_dir, 'srvkey.pem')
+            shutil.copy(ssl_key, keyfile)
+            os.chmod(keyfile, 0o400)
+            server_settings = server_settings.copy()
+            server_settings['ssl_key_file'] = keyfile
+
         if self._pg_version < (9, 3):
             sockdir_opt = 'unix_socket_directory'
         else:
@@ -204,10 +214,15 @@ class Cluster:
                     'pg_ctl start exited with status {:d}: {}'.format(
                         process.returncode, stderr.decode()))
         else:
+            if os.getenv('ASYNCPG_DEBUG_SERVER'):
+                stdout = sys.stdout
+            else:
+                stdout = subprocess.DEVNULL
+
             self._daemon_process = \
                 subprocess.Popen(
                     [self._postgres, '-D', self._data_dir, *extra_args],
-                    stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
+                    stdout=stdout, stderr=subprocess.STDOUT,
                     preexec_fn=ensure_dead_with_parent)
 
             self._daemon_pid = self._daemon_process.pid
