@@ -22,6 +22,7 @@ extensions = [
     'sphinx.ext.doctest',
     'sphinx.ext.viewcode',
     'sphinx.ext.githubpages',
+    'sphinx.ext.intersphinx',
     'sphinxcontrib.asyncio',
 ]
 
@@ -31,7 +32,7 @@ templates_path = ['_templates']
 source_suffix = '.rst'
 master_doc = 'index'
 project = 'asyncpg'
-copyright = '2016-present, the ayncpg authors and contributors'
+copyright = '2016-present, the asyncpg authors and contributors'
 author = '<See AUTHORS file>'
 release = version
 language = None
@@ -51,7 +52,7 @@ html_theme = 'sphinx_rtd_theme'
 html_theme_path = [alabaster.get_path()]
 html_title = 'asyncpg Documentation'
 html_short_title = 'asyncpg'
-html_static_path = []
+html_static_path = ['_static']
 html_sidebars = {
     '**': [
         'about.html',
@@ -61,6 +62,11 @@ html_sidebars = {
 html_show_sourcelink = False
 html_show_sphinx = False
 html_show_copyright = True
+html_context = {
+    'css_files': [
+        '_static/theme_overrides.css',
+    ],
+}
 htmlhelp_basename = 'asyncpgdoc'
 
 
@@ -91,3 +97,7 @@ texinfo_documents = [
      'Python asyncio framework',
      'Miscellaneous'),
 ]
+
+# -- Options for intersphinx ----------------------------------------------
+
+intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}
