@@ -1,4 +1,4 @@
-# Copyright (C) 2016-present the ayncpg authors and contributors
+# Copyright (C) 2016-present the asyncpg authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of asyncpg and is released under
@@ -37,15 +37,15 @@ type_samples = [
         True, False,
     )),
     ('smallint', 'int2', (
-        -2 ** 15 + 1, 2 ** 15 - 1,
+        -2 ** 15, 2 ** 15 - 1,
         -1, 0, 1,
     )),
     ('int', 'int4', (
-        -2 ** 31 + 1, 2 ** 31 - 1,
+        -2 ** 31, 2 ** 31 - 1,
         -1, 0, 1,
     )),
     ('bigint', 'int8', (
-        -2 ** 63 + 1, 2 ** 63 - 1,
+        -2 ** 63, 2 ** 63 - 1,
         -1, 0, 1,
     )),
     ('numeric', 'numeric', (
@@ -258,7 +258,10 @@ type_samples = [
         math.e,
         math.inf,
         -math.inf,
-        math.nan
+        math.nan,
+        {'textinput': 'infinity', 'output': math.inf},
+        {'textinput': '-infinity', 'output': -math.inf},
+        {'textinput': 'NaN', 'output': math.nan},
     ]),
     ('float4', 'float4', [
         1.1,
@@ -272,7 +275,10 @@ type_samples = [
         math.e,
         math.inf,
         -math.inf,
-        math.nan
+        math.nan,
+        {'textinput': 'infinity', 'output': math.inf},
+        {'textinput': '-infinity', 'output': -math.inf},
+        {'textinput': 'NaN', 'output': math.nan},
     ]),
     ('cidr', 'cidr', [
         ipaddress.IPv4Network('255.255.255.255/32'),
@@ -433,6 +439,56 @@ class TestCodecs(tb.ConnectedTestCase):
 
         self.assertEqual(len(bits.bytes), expected_bytelen)
 
+    async def test_interval(self):
+        res = await self.con.fetchval("SELECT '5 years'::interval")
+        self.assertEqual(res, datetime.timedelta(days=1825))
+
+        res = await self.con.fetchval("SELECT '5 years 1 month'::interval")
+        self.assertEqual(res, datetime.timedelta(days=1855))
+
+        res = await self.con.fetchval("SELECT '-5 years'::interval")
+        self.assertEqual(res, datetime.timedelta(days=-1825))
+
+        res = await self.con.fetchval("SELECT '-5 years -1 month'::interval")
+        self.assertEqual(res, datetime.timedelta(days=-1855))
+
+    async def test_numeric(self):
+        # Test that we handle dscale correctly.
+        cases = [
+            '0.001',
+            '0.001000',
+            '1',
+            '1.00000'
+        ]
+
+        for case in cases:
+            res = await self.con.fetchval(
+                "SELECT $1::numeric", case)
+
+            self.assertEqual(str(res), case)
+
+        res = await self.con.fetchval(
+            "SELECT $1::numeric", decimal.Decimal('NaN'))
+        self.assertTrue(res.is_nan())
+
+        res = await self.con.fetchval(
+            "SELECT $1::numeric", decimal.Decimal('sNaN'))
+        self.assertTrue(res.is_nan())
+
+        with self.assertRaisesRegex(ValueError, 'numeric type does not '
+                                                'support infinite values'):
+            await self.con.fetchval(
+                "SELECT $1::numeric", decimal.Decimal('-Inf'))
+
+        with self.assertRaisesRegex(ValueError, 'numeric type does not '
+                                                'support infinite values'):
+            await self.con.fetchval(
+                "SELECT $1::numeric", decimal.Decimal('+Inf'))
+
+        with self.assertRaises(decimal.InvalidOperation):
+            await self.con.fetchval(
+                "SELECT $1::numeric", 'invalid')
+
     async def test_unhandled_type_fallback(self):
         await self.con.execute('''
             CREATE EXTENSION IF NOT EXISTS isn
@@ -466,9 +522,13 @@ class TestCodecs(tb.ConnectedTestCase):
                 '2',
                 'aa',
             ]),
-            ('smallint', ValueError, 'integer too large', [
+            ('smallint', OverflowError, 'int too big to be encoded as INT2', [
+                2**256,  # check for the same exception for any big numbers
+                decimal.Decimal("2000000000000000000000000000000"),
+                0xffff,
+                0xffffffff,
                 32768,
-                -32768
+                -32769
             ]),
             ('float4', ValueError, 'float value too large', [
                 4.1 * 10 ** 40,
@@ -478,9 +538,29 @@ class TestCodecs(tb.ConnectedTestCase):
                 '2',
                 'aa',
             ]),
+            ('int', OverflowError, 'int too big to be encoded as INT4', [
+                2**256,  # check for the same exception for any big numbers
+                decimal.Decimal("2000000000000000000000000000000"),
+                0xffffffff,
+                2**31,
+                -2**31 - 1,
+            ]),
             ('int8', TypeError, 'an integer is required', [
                 '2',
                 'aa',
+            ]),
+            ('bigint', OverflowError, 'int too big to be encoded as INT8', [
+                2**256,  # check for the same exception for any big numbers
+                decimal.Decimal("2000000000000000000000000000000"),
+                0xffffffffffffffff,
+                2**63,
+                -2**63 - 1,
+            ]),
+            ('text', TypeError, 'expected str, got bytes', [
+                b'foo'
+            ]),
+            ('text', TypeError, 'expected str, got list', [
+                [1]
             ])
         ]
 
@@ -507,6 +587,10 @@ class TestCodecs(tb.ConnectedTestCase):
                 r"SELECT '{1, 2, NULL}'::int[]::anyarray",
                 [1, 2, None]
             ),
+            (
+                r"SELECT '{}'::int[]",
+                []
+            ),
         ]
 
         for sql, expected in cases:
@@ -725,7 +809,7 @@ class TestCodecs(tb.ConnectedTestCase):
 
             res = await self.con.fetchval('''
                 SELECT $1::hstore AS result
-            ''', (('foo', 2), ('bar', 3)))
+            ''', (('foo', '2'), ('bar', '3')))
 
             self.assertEqual(res, {'foo': '2', 'bar': '3'})
 
@@ -861,7 +945,7 @@ class TestCodecs(tb.ConnectedTestCase):
         try:
             await self.con.set_type_codec('hstore', encoder=hstore_encoder,
                                           decoder=hstore_decoder,
-                                          binary=True)
+                                          format='binary')
 
             st = await self.con.prepare('''
                 SELECT $1::hstore AS result
@@ -890,6 +974,152 @@ class TestCodecs(tb.ConnectedTestCase):
                 DROP EXTENSION hstore
             ''')
 
+    async def test_custom_codec_override_binary(self):
+        """Test overriding core codecs."""
+        import json
+
+        conn = await self.cluster.connect(database='postgres', loop=self.loop)
+        try:
+            def _encoder(value):
+                return json.dumps(value).encode('utf-8')
+
+            def _decoder(value):
+                return json.loads(value.decode('utf-8'))
+
+            await conn.set_type_codec(
+                'json', encoder=_encoder, decoder=_decoder,
+                schema='pg_catalog', format='binary'
+            )
+
+            data = {'foo': 'bar', 'spam': 1}
+            res = await conn.fetchval('SELECT $1::json', data)
+            self.assertEqual(data, res)
+
+        finally:
+            await conn.close()
+
+    async def test_custom_codec_override_text(self):
+        """Test overriding core codecs."""
+        import json
+
+        conn = await self.cluster.connect(database='postgres', loop=self.loop)
+        try:
+            def _encoder(value):
+                return json.dumps(value)
+
+            def _decoder(value):
+                return json.loads(value)
+
+            await conn.set_type_codec(
+                'json', encoder=_encoder, decoder=_decoder,
+                schema='pg_catalog', format='text'
+            )
+
+            data = {'foo': 'bar', 'spam': 1}
+            res = await conn.fetchval('SELECT $1::json', data)
+            self.assertEqual(data, res)
+
+            def _encoder(value):
+                return value
+
+            def _decoder(value):
+                return value
+
+            await conn.set_type_codec(
+                'uuid', encoder=_encoder, decoder=_decoder,
+                schema='pg_catalog', format='text'
+            )
+
+            data = '14058ad9-0118-4b7e-ac15-01bc13e2ccd1'
+            res = await conn.fetchval('SELECT $1::uuid', data)
+            self.assertEqual(res, data)
+        finally:
+            await conn.close()
+
+    async def test_custom_codec_override_tuple(self):
+        """Test overriding core codecs."""
+        cases = [
+            ('date', (3,), '2000-01-04'),
+            ('date', (2**31 - 1,), 'infinity'),
+            ('date', (-2**31,), '-infinity'),
+            ('time', (60 * 10**6,), '00:01:00'),
+            ('timetz', (60 * 10**6, 12600), '00:01:00-03:30'),
+            ('timestamp', (60 * 10**6,), '2000-01-01 00:01:00'),
+            ('timestamp', (2**63 - 1,), 'infinity'),
+            ('timestamp', (-2**63,), '-infinity'),
+            ('timestamptz', (60 * 10**6,), '1999-12-31 19:01:00',
+                "tab.v AT TIME ZONE 'EST'"),
+            ('timestamptz', (2**63 - 1,), 'infinity'),
+            ('timestamptz', (-2**63,), '-infinity'),
+            ('interval', (2, 3, 1), '2 mons 3 days 00:00:00.000001')
+        ]
+
+        conn = await self.cluster.connect(database='postgres', loop=self.loop)
+
+        def _encoder(value):
+            return tuple(value)
+
+        def _decoder(value):
+            return tuple(value)
+
+        try:
+            for (typename, data, expected_result, *extra) in cases:
+                with self.subTest(type=typename):
+                    await self.con.execute(
+                        'CREATE TABLE tab (v {})'.format(typename))
+
+                    try:
+                        await conn.set_type_codec(
+                            typename, encoder=_encoder, decoder=_decoder,
+                            schema='pg_catalog', format='tuple'
+                        )
+
+                        await conn.execute(
+                            'INSERT INTO tab VALUES ($1)', data)
+
+                        res = await conn.fetchval('SELECT tab.v FROM tab')
+                        self.assertEqual(res, data)
+
+                        await conn.reset_type_codec(
+                            typename, schema='pg_catalog')
+
+                        if extra:
+                            val = extra[0]
+                        else:
+                            val = 'tab.v'
+
+                        res = await conn.fetchval(
+                            'SELECT ({val})::text FROM tab'.format(val=val))
+                        self.assertEqual(res, expected_result)
+                    finally:
+                        await self.con.execute('DROP TABLE tab')
+        finally:
+            await conn.close()
+
+    async def test_custom_codec_override_deprecation(self):
+        conn = await self.cluster.connect(database='postgres', loop=self.loop)
+        try:
+            def _encoder(value):
+                return value
+
+            def _decoder(value):
+                return value
+
+            with self.assertWarnsRegex(DeprecationWarning,
+                                       r"The `binary` keyword argument to "
+                                       r"set_type_codec\(\) is deprecated"):
+                await conn.set_type_codec(
+                    'uuid', encoder=_encoder, decoder=_decoder,
+                    schema='pg_catalog', binary=False
+                )
+
+                data = '14058ad9-0118-4b7e-ac15-01bc13e2ccd1'
+                res = await conn.fetchval('SELECT $1::uuid', data)
+                self.assertEqual(res, data)
+
+        finally:
+            await conn.close()
+
     async def test_composites_in_arrays(self):
         await self.con.execute('''
             CREATE TYPE t AS (a text, b int);
@@ -1018,3 +1248,157 @@ class TestCodecs(tb.ConnectedTestCase):
                 DROP USER norm1;
                 DROP USER norm2;
             ''')
+
+    async def test_enum(self):
+        await self.con.execute('''
+            CREATE TYPE enum_t AS ENUM ('abc', 'def', 'ghi');
+            CREATE TABLE tab (
+                a text,
+                b enum_t
+            );
+            INSERT INTO tab (a, b) VALUES ('foo', 'abc');
+            INSERT INTO tab (a, b) VALUES ('bar', 'def');
+        ''')
+
+        try:
+            for i in range(10):
+                r = await self.con.fetch('''
+                    SELECT a, b FROM tab ORDER BY b
+                ''')
+
+                self.assertEqual(r, [('foo', 'abc'), ('bar', 'def')])
+
+        finally:
+            await self.con.execute('''
+                DROP TABLE tab;
+                DROP TYPE enum_t;
+            ''')
+
+    async def test_unknown_type_text_fallback(self):
+        await self.con.execute(r'CREATE EXTENSION citext')
+        await self.con.execute(r'''
+            CREATE DOMAIN citext_dom AS citext
+        ''')
+        await self.con.execute(r'''
+            CREATE TYPE citext_range AS RANGE (SUBTYPE = citext)
+        ''')
+        await self.con.execute(r'''
+            CREATE TYPE citext_comp AS (t citext)
+        ''')
+
+        try:
+            # Check that plain fallback works.
+            result = await self.con.fetchval('''
+                SELECT $1::citext
+            ''', 'citext')
+
+            self.assertEqual(result, 'citext')
+
+            # Check that domain fallback works.
+            result = await self.con.fetchval('''
+                SELECT $1::citext_dom
+            ''', 'citext')
+
+            self.assertEqual(result, 'citext')
+
+            # Check that array fallback works.
+            cases = [
+                ['a', 'b'],
+                [None, 'b'],
+                [],
+                ['  a', '  b'],
+                ['"a', r'\""'],
+                [['"a', r'\""'], [',', '",']],
+            ]
+
+            for case in cases:
+                result = await self.con.fetchval('''
+                    SELECT
+                        $1::citext[]
+                ''', case)
+
+                self.assertEqual(result, case)
+
+            # Text encoding of ranges and composite types
+            # is not supported yet.
+            with self.assertRaisesRegex(
+                    RuntimeError,
+                    'text encoding of range types is not supported'):
+
+                await self.con.fetchval('''
+                    SELECT
+                        $1::citext_range
+                ''', ['a', 'z'])
+
+            with self.assertRaisesRegex(
+                    RuntimeError,
+                    'text encoding of composite types is not supported'):
+
+                await self.con.fetchval('''
+                    SELECT
+                        $1::citext_comp
+                ''', ('a',))
+
+            # Check that setting a custom codec clears the codec
+            # cache properly and that subsequent queries work
+            # as expected.
+            await self.con.set_type_codec(
+                'citext', encoder=lambda d: d, decoder=lambda d: 'CI: ' + d)
+
+            result = await self.con.fetchval('''
+                SELECT
+                    $1::citext[]
+            ''', ['a', 'b'])
+
+            self.assertEqual(result, ['CI: a', 'CI: b'])
+
+        finally:
+            await self.con.execute(r'DROP TYPE citext_comp')
+            await self.con.execute(r'DROP TYPE citext_range')
+            await self.con.execute(r'DROP TYPE citext_dom')
+            await self.con.execute(r'DROP EXTENSION citext')
+
+    async def test_enum_in_array(self):
+        await self.con.execute('''
+            CREATE TYPE enum_t AS ENUM ('abc', 'def', 'ghi');
+        ''')
+
+        try:
+            result = await self.con.fetchrow('''SELECT $1::enum_t[];''',
+                                             ['abc'])
+            self.assertEqual(result, (['abc'],))
+
+            result = await self.con.fetchrow('''SELECT ARRAY[$1::enum_t];''',
+                                             'abc')
+
+            self.assertEqual(result, (['abc'],))
+
+        finally:
+            await self.con.execute('''
+                DROP TYPE enum_t;
+            ''')
+
+    async def test_enum_and_range(self):
+        await self.con.execute('''
+            CREATE TYPE enum_t AS ENUM ('abc', 'def', 'ghi');
+            CREATE TABLE testtab (
+                a int4range,
+                b enum_t
+            );
+
+            INSERT INTO testtab VALUES (
+                '[10, 20)', 'abc'
+            );
+        ''')
+
+        try:
+            result = await self.con.fetchrow('''
+                SELECT testtab.a FROM testtab WHERE testtab.b = $1
+            ''', 'abc')
+
+            self.assertEqual(result, (asyncpg.Range(10, 20),))
+        finally:
+            await self.con.execute('''
+                DROP TABLE testtab;
+                DROP TYPE enum_t;
+            ''')
