@@ -1,4 +1,4 @@
-# Copyright (C) 2016-present the ayncpg authors and contributors
+# Copyright (C) 2016-present the asyncpg authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of asyncpg and is released under
@@ -142,12 +142,14 @@ class TestPrepare(tb.ConnectedTestCase):
                 await stmt.fetchval()
 
     async def test_prepare_10_stmt_lru(self):
+        cache = self.con._stmt_cache
+
         query = 'select {}'
-        cache_max = self.con._stmt_cache_max_size
+        cache_max = cache.get_max_size()
         iter_max = cache_max * 2 + 11
 
         # First, we have no cached statements.
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        self.assertEqual(len(cache), 0)
 
         stmts = []
         for i in range(iter_max):
@@ -156,9 +158,8 @@ class TestPrepare(tb.ConnectedTestCase):
             stmts.append(s)
 
         # At this point our cache should be full.
-        self.assertEqual(len(self.con._stmt_cache), cache_max)
-        self.assertTrue(
-            all(not s.closed for s in self.con._stmt_cache.values()))
+        self.assertEqual(len(cache), cache_max)
+        self.assertTrue(all(not s.closed for s in cache.iter_statements()))
 
         # Since there are references to the statements (`stmts` list),
         # no statements are scheduled to be closed.
@@ -173,22 +174,20 @@ class TestPrepare(tb.ConnectedTestCase):
         # scheduled to be closed.
         self.assertEqual(len(self.con._stmts_to_close), iter_max - cache_max)
         self.assertTrue(all(s.closed for s in self.con._stmts_to_close))
-        self.assertTrue(
-            all(not s.closed for s in self.con._stmt_cache.values()))
+        self.assertTrue(all(not s.closed for s in cache.iter_statements()))
 
         zero = await self.con.prepare(query.format(0))
         # Hence, all stale statements should be closed now.
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
         # The number of cached statements will stay the same though.
-        self.assertEqual(len(self.con._stmt_cache), cache_max)
-        self.assertTrue(
-            all(not s.closed for s in self.con._stmt_cache.values()))
+        self.assertEqual(len(cache), cache_max)
+        self.assertTrue(all(not s.closed for s in cache.iter_statements()))
 
         # After closing all statements will be closed.
         await self.con.close()
         self.assertEqual(len(self.con._stmts_to_close), 0)
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        self.assertEqual(len(cache), 0)
 
         # An attempt to perform an operation on a closed statement
         # will trigger an error.
@@ -199,8 +198,10 @@ class TestPrepare(tb.ConnectedTestCase):
         # Test that prepared statements should stay in the cache after
         # they are GCed.
 
+        cache = self.con._stmt_cache
+
         # First, we have no cached statements.
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        self.assertEqual(len(cache), 0)
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
         # The prepared statement that we'll create will be GCed
@@ -209,33 +210,34 @@ class TestPrepare(tb.ConnectedTestCase):
         await self.con.prepare('select 1')
         gc.collect()
 
-        self.assertEqual(len(self.con._stmt_cache), 1)
+        self.assertEqual(len(cache), 1)
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
     async def test_prepare_12_stmt_gc(self):
         # Test that prepared statements are closed when there is no space
         # for them in the LRU cache and there are no references to them.
 
+        cache = self.con._stmt_cache
+        cache_max = cache.get_max_size()
+
         # First, we have no cached statements.
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        self.assertEqual(len(cache), 0)
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
-        cache_max = self.con._stmt_cache_max_size
-
         stmt = await self.con.prepare('select 100000000')
-        self.assertEqual(len(self.con._stmt_cache), 1)
+        self.assertEqual(len(cache), 1)
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
         for i in range(cache_max):
             await self.con.prepare('select {}'.format(i))
 
-        self.assertEqual(len(self.con._stmt_cache), cache_max)
+        self.assertEqual(len(cache), cache_max)
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
         del stmt
         gc.collect()
 
-        self.assertEqual(len(self.con._stmt_cache), cache_max)
+        self.assertEqual(len(cache), cache_max)
         self.assertEqual(len(self.con._stmts_to_close), 1)
 
     async def test_prepare_13_connect(self):
@@ -283,25 +285,28 @@ class TestPrepare(tb.ConnectedTestCase):
         # Test that even if the statements cache is off, we're still
         # cleaning up GCed statements.
 
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        cache = self.con._stmt_cache
+
+        self.assertEqual(len(cache), 0)
         self.assertEqual(len(self.con._stmts_to_close), 0)
+
         # Disable cache
-        self.con._stmt_cache_max_size = 0
+        cache.set_max_size(0)
 
         stmt = await self.con.prepare('select 100000000')
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        self.assertEqual(len(cache), 0)
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
         del stmt
         gc.collect()
 
         # After GC, _stmts_to_close should contain stmt's state
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        self.assertEqual(len(cache), 0)
         self.assertEqual(len(self.con._stmts_to_close), 1)
 
         # Next "prepare" call will trigger a cleanup
         stmt = await self.con.prepare('select 1')
-        self.assertEqual(len(self.con._stmt_cache), 0)
+        self.assertEqual(len(cache), 0)
         self.assertEqual(len(self.con._stmts_to_close), 0)
 
         del stmt
@@ -355,7 +360,7 @@ class TestPrepare(tb.ConnectedTestCase):
 
     async def test_prepare_19_concurrent_calls(self):
         st = self.loop.create_task(self.con.fetchval(
-            'SELECT ROW(pg_sleep(0.03), 1)'))
+            'SELECT ROW(pg_sleep(0.1), 1)'))
 
         # Wait for some time to make sure the first query is fully
         # prepared (!) and is now awaiting the results (!!).
@@ -379,7 +384,7 @@ class TestPrepare(tb.ConnectedTestCase):
                 meth = getattr(self.con, methname)
 
                 vf = self.loop.create_task(
-                    meth('SELECT ROW(pg_sleep(0.03), 1)'))
+                    meth('SELECT ROW(pg_sleep(0.1), 1)'))
 
                 await asyncio.sleep(0.01, loop=self.loop)
 
@@ -406,3 +411,148 @@ class TestPrepare(tb.ConnectedTestCase):
         result = await self.con.fetchrow('SELECT')
         self.assertEqual(result, ())
         self.assertEqual(repr(result), '<Record>')
+
+    async def test_prepare_statement_invalid(self):
+        await self.con.execute('CREATE TABLE tab1(a int, b int)')
+
+        try:
+            await self.con.execute('INSERT INTO tab1 VALUES (1, 2)')
+
+            stmt = await self.con.prepare('SELECT * FROM tab1')
+
+            await self.con.execute(
+                'ALTER TABLE tab1 ALTER COLUMN b SET DATA TYPE text')
+
+            with self.assertRaisesRegex(asyncpg.InvalidCachedStatementError,
+                                        'cached statement plan is invalid'):
+                await stmt.fetchrow()
+
+        finally:
+            await self.con.execute('DROP TABLE tab1')
+
+    @tb.with_connection_options(statement_cache_size=0)
+    async def test_prepare_23_no_stmt_cache_seq(self):
+        self.assertEqual(self.con._stmt_cache.get_max_size(), 0)
+
+        async def check_simple():
+            # Run a simple query a few times.
+            self.assertEqual(await self.con.fetchval('SELECT 1'), 1)
+            self.assertEqual(await self.con.fetchval('SELECT 2'), 2)
+            self.assertEqual(await self.con.fetchval('SELECT 1'), 1)
+
+        await check_simple()
+
+        # Run a query that timeouts.
+        with self.assertRaises(asyncio.TimeoutError):
+            await self.con.fetchrow('select pg_sleep(10)', timeout=0.02)
+
+        # Check that we can run new queries after a timeout.
+        await check_simple()
+
+        # Try a cursor/timeout combination. Cursors should always use
+        # named prepared statements.
+        async with self.con.transaction():
+            with self.assertRaises(asyncio.TimeoutError):
+                async for _ in self.con.cursor(   # NOQA
+                        'select pg_sleep(10)', timeout=0.1):
+                    pass
+
+        # Check that we can run queries after a failed cursor
+        # operation.
+        await check_simple()
+
+    @tb.with_connection_options(max_cached_statement_lifetime=142)
+    async def test_prepare_24_max_lifetime(self):
+        cache = self.con._stmt_cache
+
+        self.assertEqual(cache.get_max_lifetime(), 142)
+        cache.set_max_lifetime(1)
+
+        s = await self.con.prepare('SELECT 1')
+        state = s._state
+
+        s = await self.con.prepare('SELECT 1')
+        self.assertIs(s._state, state)
+
+        s = await self.con.prepare('SELECT 1')
+        self.assertIs(s._state, state)
+
+        await asyncio.sleep(1, loop=self.loop)
+
+        s = await self.con.prepare('SELECT 1')
+        self.assertIsNot(s._state, state)
+
+    @tb.with_connection_options(max_cached_statement_lifetime=0.5)
+    async def test_prepare_25_max_lifetime_reset(self):
+        cache = self.con._stmt_cache
+
+        s = await self.con.prepare('SELECT 1')
+        state = s._state
+
+        # Disable max_lifetime
+        cache.set_max_lifetime(0)
+
+        await asyncio.sleep(1, loop=self.loop)
+
+        # The statement should still be cached (as we disabled the timeout).
+        s = await self.con.prepare('SELECT 1')
+        self.assertIs(s._state, state)
+
+    @tb.with_connection_options(max_cached_statement_lifetime=0.5)
+    async def test_prepare_26_max_lifetime_max_size(self):
+        cache = self.con._stmt_cache
+
+        s = await self.con.prepare('SELECT 1')
+        state = s._state
+
+        # Disable max_lifetime
+        cache.set_max_size(0)
+
+        s = await self.con.prepare('SELECT 1')
+        self.assertIsNot(s._state, state)
+
+        # Check that nothing crashes after the initial timeout
+        await asyncio.sleep(1, loop=self.loop)
+
+    @tb.with_connection_options(max_cacheable_statement_size=50)
+    async def test_prepare_27_max_cacheable_statement_size(self):
+        cache = self.con._stmt_cache
+
+        await self.con.prepare('SELECT 1')
+        self.assertEqual(len(cache), 1)
+
+        # Test that long and explicitly created prepared statements
+        # are not cached.
+        await self.con.prepare("SELECT \'" + "a" * 50 + "\'")
+        self.assertEqual(len(cache), 1)
+
+        # Test that implicitly created long prepared statements
+        # are not cached.
+        await self.con.fetchval("SELECT \'" + "a" * 50 + "\'")
+        self.assertEqual(len(cache), 1)
+
+        # Test that short prepared statements can still be cached.
+        await self.con.prepare('SELECT 2')
+        self.assertEqual(len(cache), 2)
+
+    async def test_prepare_28_max_args(self):
+        N = 32768
+        args = ','.join('${}'.format(i) for i in range(1, N + 1))
+        query = 'SELECT ARRAY[{}]'.format(args)
+
+        with self.assertRaisesRegex(ValueError,
+                                    'number of arguments cannot exceed 32767'):
+            await self.con.fetchval(query, *range(1, N + 1))
+
+    async def test_prepare_29_duplicates(self):
+        # In addition to test_record.py, let's have a full functional
+        # test for records with duplicate keys.
+        r = await self.con.fetchrow('SELECT 1 as a, 2 as b, 3 as a')
+        self.assertEqual(list(r.items()), [('a', 1), ('b', 2), ('a', 3)])
+        self.assertEqual(list(r.keys()), ['a', 'b', 'a'])
+        self.assertEqual(list(r.values()), [1, 2, 3])
+        self.assertEqual(r['a'], 3)
+        self.assertEqual(r['b'], 2)
+        self.assertEqual(r[0], 1)
+        self.assertEqual(r[1], 2)
+        self.assertEqual(r[2], 3)
